#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "kikstart.h"

#include "keyboard.h"
#include "cpu.h"
#include "tedmem.h"
#include "tape.h"
#include "sound.h"

#include "gpsdk/gpsdk.h"

u8 P4Colors[128][3] =
	{
		/* Black */
		{ 0,   0,   0   },
		{ 0,   0,   0   },
		{ 0,   0,   0   },
		{ 0,   0,   0   },
		{ 0,   0,   0   },
		{ 0,   0,   0   },
		{ 0,   0,   0   },
		{ 0,   0,   0   },
		/* White */
		{ 42,  42,  42  },
		{ 56,  56,  56  },
		{ 70,  70,  70  },
		{ 87,  87,  87  },
		{ 121, 121, 121 },
		{ 146, 146, 146 },
		{ 185, 185, 185 },
		{ 239, 239, 239 },
		/* Red */
		{ 103, 12,  13  },
		{ 116, 28,  29  },
		{ 129, 43,  44  },
		{ 145, 61,  62  },
		{ 177, 96,  97  },
		{ 200, 121, 122 },
		{ 239, 162, 163 },
		{ 255, 216, 217 },
		/* Cyan */
		{ 0,   70,  69  },
		{ 0,   82,  81  },
		{ 0,   95,  94  },
		{ 23,  112, 111 },
		{ 61,  145, 144 },
		{ 88,  170, 169 },
		{ 130, 208, 207 },
		{ 186, 255, 255 },
		/* Magenta */
		{ 94,  0,   91  },
		{ 106, 17,  104 },
		{ 119, 33,  117 },
		{ 136, 52,  134 },
		{ 168, 87,  166 },
		{ 192, 112, 189 },
		{ 230, 154, 228 },
		{ 255, 208, 255 },
		/* Green */
		{ 0,   79,  0   },
		{ 0,   92,  12  },
		{ 8,   105, 28  },
		{ 29,  122, 47  },
		{ 67,  155, 83  },
		{ 93,  178, 109 },
		{ 135, 217, 150 },
		{ 191, 255, 205 },
		/* Blue */
		{ 42,  24,  122 },
		{ 56,  39,  135 },
		{ 70,  54,  147 },
		{ 87,  72,  163 },
		{ 121, 106, 195 },
		{ 145, 131, 217 },
		{ 185, 171, 255 },
		{ 239, 225, 255 },
		/* Yellow */
		{ 43,  59,  0   },
		{ 56,  72,  0   },
		{ 70,  85,  0   },
		{ 87,  102, 0   },
		{ 121, 136, 41  },
		{ 146, 160, 69  },
		{ 185, 199, 111 },
		{ 239, 253, 169 },
		/* Orange */
		{ 96,  21,  0   },
		{ 109, 36,  6   },
		{ 122, 51,  23  },
		{ 139, 69,  43  },
		{ 171, 103, 79  },
		{ 194, 128, 104 },
		{ 232, 169, 146 },
		{ 255, 223, 201 },
		/* Brown */
		{ 77,  38,  0   },
		{ 90,  52,  0   },
		{ 103, 66,  0   },
		{ 120, 83,  14  },
		{ 153, 117, 54  },
		{ 176, 142, 80  },
		{ 215, 182, 123 },
		{ 255, 236, 179 },
		/* Yllw-Grn */
		{ 24,  67,  0   },
		{ 38,  80,  0   },
		{ 53,  93,  0   },
		{ 71,  110, 1   },
		{ 105, 143, 44  },
		{ 131, 167, 71  },
		{ 171, 206, 114 },
		{ 225, 255, 171 },
		/* Pink */
		{ 103, 0,   67  },
		{ 116, 17,  80  },
		{ 129, 33,  93  },
		{ 145, 52,  110 },
		{ 177, 87,  143 },
		{ 200, 112, 167 },
		{ 239, 154, 206 },
		{ 255, 208, 255 },
		/* Blu-Grn */
		{ 0,   77,  35  },
		{ 0,   90,  48  },
		{ 0,   103, 62  },
		{ 21,  120, 80  },
		{ 59,  153, 114 },
		{ 85,  176, 139 },
		{ 128, 215, 179 },
		{ 184, 255, 233 },
		/* Lt-Blu */
		{ 0,   49,  107 },
		{ 13,  63,  119 },
		{ 29,  77,  132 },
		{ 48,  94,  148 },
		{ 84,  128, 180 },
		{ 109, 152, 204 },
		{ 151, 192, 241 },
		{ 205, 245, 255 },
		/* Dk-Blu */
		{ 51,  21,  122 },
		{ 65,  35,  135 },
		{ 77,  49,  146 },
		{ 94,  68,  162 },
		{ 128, 102, 194 },
		{ 153, 128, 217 },
		{ 192, 168, 255 },
		{ 246, 223, 255 },
		/* Lt-Grn */
		{ 0,   74,  0   },
		{ 18,  87,  0   },
		{ 33,  100, 0   },
		{ 52,  117, 16  },
		{ 87,  150, 55  },
		{ 113, 173, 81  },
		{ 154, 213, 124 },
		{ 209, 255, 180 }
	};

/* Converts 8bit rgb values to a GP32 palette value */
#define GP_RGB24(r,g,b) (((((r>>3))&0x1f)<<11)|((((g>>3))&0x1f)<<6)|((((b>>3))&0x1f)<<1))

/* Global variables */
GPDRAWSURFACE gpDraw[2];

/* Sets a single GP32 palette entry */
void GpSetPaletteEntry ( u8 i, u8 r, u8 g, u8 b )
  {
    GP_PALETTEENTRY entry = GP_RGB24(r,g,b);
    GpPaletteEntryChange ( i, 1, &entry, 0 );
  }

static int filesize ( char * fname )
  {
    FILE * f;
    int s;

    f = fopen (fname,"rb");
    if (f==NULL)
      return -1;

    fseek(f, 0, SEEK_END);
    s=ftell(f);
    fclose(f);

    return s;
  }

bool load_file(CPU * machine, char *fname)
{
	u8   *lpBufPtr;
	u16	loadaddr;
	FILE	*prg;
	u32	fsize;

        fsize = 9559;
/*        fsize = filesize ( fname );
	if ((prg = fopen(fname, "rb"))== NULL) {
    	return false;
	} else*/ {
                int i;
		// load PRG file
		lpBufPtr=/*(u8 *) malloc(fsize)*/kikstart;
/*		fread(lpBufPtr,fsize,1,prg);*/
		// copy to memory
		loadaddr=lpBufPtr[0]|(lpBufPtr[1]<<8);
		for (i=0;i<fsize;i++)
			machine->mem->wrt(loadaddr+i,lpBufPtr[2+i]);

		machine->mem->wrt(0x2D,(loadaddr+fsize-2)&0xFF);
		machine->mem->wrt(0x2E,(loadaddr+fsize-2)>>8);
		machine->mem->wrt(0x2F,(loadaddr+fsize-2)&0xFF);
		machine->mem->wrt(0x30,(loadaddr+fsize-2)>>8);
		machine->mem->wrt(0x31,(loadaddr+fsize-2)&0xFF);
		machine->mem->wrt(0x32,(loadaddr+fsize-2)>>8);
		machine->mem->wrt(0x9D,(loadaddr+fsize-2)&0xFF);
		machine->mem->wrt(0x9E,(loadaddr+fsize-2)>>8);

/*		fclose(prg);*/
	};

	return true;
}

bool StartFile(CPU * machine)
{
	machine->mem->wrt(0xEF,4);
	machine->mem->wrt(0x0527,82);
	machine->mem->wrt(0x0528,85);
	machine->mem->wrt(0x0529,78);
	machine->mem->wrt(0x052A,13);
	return true;
}

void SetP4Palette ( int alpha )
  {
    int i;
    for (i=0;i<128;i++)
      {
        u8 c = ((i&0x0f)<<3)|((i>>4)&0x07);
        u32 r = P4Colors[c][0];
        u32 g = P4Colors[c][1];
        u32 b = P4Colors[c][2];

        r = (r*alpha)>>4;
        g = (g*alpha)>>4;
        b = (b*alpha)>>4;
        r = (r>255 ? 255 : r);
        g = (g>255 ? 255 : g);
        b = (b>255 ? 255 : b);
        GpSetPaletteEntry ( i, r, g, b );
      }
  }

void GpMain (void * arg)
  {
    int * t;
    int i;
    u8 * screenbuf = (u8*)malloc ( 320*240 );
    CPU * machine = new CPU;

    memset(screenbuf, 0, 320*240);

    /* Set speed to 132MHz and switch cache on */
    GpClockSpeedChange (132000000, 0x3a011, 3);
    EnableCache ();

    GpFatInit();
    GpKeyInit();

    /* Initialize graphics */
    GpGraphicModeSet(8, t);
    GpLcdSurfaceGet(&gpDraw[0], 0);
    GpLcdSurfaceGet(&gpDraw[1], 1);
    GpSurfaceSet(&gpDraw[0]);
    GpLcdEnable();

    GpSurfaceFlip(&gpDraw[0]);
    SetP4Palette ( 24 );

    machine->mem->joyemu = 1;
    machine->mem->keys->joyinit();

    while (true)
      {
        machine->mem->ted_process();
        if (machine->mem->render_ok)
          {
            int x,y1,y2,k;
            for ( y1=0,y2=(56*456)+68;y1<(200*320);y1+=320,y2+=456)
              for (x=0;x<320;++x)
                screenbuf[y1+x] = machine->mem->screen[y2+x];
            XlatVGABuffer ( screenbuf, gpDraw[0].ptbuffer );
            if (GpKeyGet()&GPC_VK_FL)
              {
                if (load_file(machine, "gp:\\gpmm\\p4emu\\kikstar.prg"))
                  StartFile(machine);
                while (GpKeyGet());
              }

            k = GpKeyGet();
            if (k&GPC_VK_UP)
              machine->mem->keys->pushkey(273);
            else
              machine->mem->keys->releasekey(273);

            if (k&GPC_VK_DOWN)
              machine->mem->keys->pushkey(274);
            else
              machine->mem->keys->releasekey(274);

            if (k&GPC_VK_RIGHT)
              machine->mem->keys->pushkey(275);
            else
              machine->mem->keys->releasekey(275);

            if (k&GPC_VK_LEFT)
              machine->mem->keys->pushkey(276);
            else
              machine->mem->keys->releasekey(276);

            if (k&GPC_VK_FA)
              machine->mem->keys->pushkey(305);
            else
              machine->mem->keys->releasekey(305);

            machine->mem->render_ok = false;
          }
      }

    GpAppExit();
  }

